@@ -67,8 +67,6 @@ readtodelim(FILE *f, const char *name, /*@unique@*/ /*@observer@*/ const char *d
 	*bufp = critrealloc(*bufp, (*size)*=2, "readtodelim");
     }
     /*@=loopexec@*/
-    /*@notreached@*/
-    return NULL;
 }
 
 /* read article headers, cut off body
@@ -137,14 +135,21 @@ main(int argc, char *argv[])
 	readfilter(filterfile);
     else {
 	printf("Nothing to filter -- no filterfile found.\n");
-	exit(1);
+	freeconfig();
+	exit(0);
     }
 
-    if (try_lock(5)) {
+    if (try_lock(timeout_lock)) {
 	printf("Cannot obtain lock file, aborting.\n");
 	exit(1);
     }
+
     readactive();
+    if (!active) {
+	printf("Problem reading active file.\n");
+	freeconfig();
+	exit(1);
+    }
 
     g = findgroup(argv[optind]);
     if (!g) {
@@ -181,7 +186,7 @@ main(int argc, char *argv[])
 	}
 	switch (verbose) {
 	case 1:{
-		printf("%c", c[i % 4]);
+		printf("%c\b", c[i % 4]);
 		fflush(stdout);
 		i++;
 		break;
@@ -265,11 +270,14 @@ main(int argc, char *argv[])
 	g->first = g->last + 1;
     }
     if (writeactive()) ln_log(LNLOG_SERR, LNLOG_CTOP, "Error writing groupinfo.");
+    freeactive(active);
     unlink(lockfile);
     printf("%d articles deleted, %d kept.\n", deleted, kept);
 
     if (verbose)
 	printf("Updating .overview file\n");
     getxover();
+    freexover();
+    freeconfig();
     exit(0);
 }
