@@ -28,20 +28,29 @@ See file COPYING for restrictions on the use of this software.
 
 #include <ctype.h>
 #include <errno.h>
+#include <sys/types.h>
 #include <limits.h>
+#include <netdb.h>
+#include <netinet/in.h>
+#ifndef __LCLINT__
+#include <arpa/inet.h>
+#endif
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <syslog.h>
 #include <sys/stat.h>
-#include <sys/types.h>
 #include <sys/resource.h>
 
-#define COREFILESIZE 1024*1024*16
+#define COREFILESIZE 1024*1024*64
 #define TOKENSIZE 4096
 
 #include "groupselect.h"
 
+#ifndef min
+#define min(a,b) ((a < b) ? (a) : (b))
+#endif
+
 /*
  * misc. global variables, documented in leafnode.h
  */
@@ -60,13 +69,14 @@ int article_despite_filter = 0;
 long maxlines = 0;
 long minlines = 0;
 unsigned long maxbytes = 0;
-int linebuffer = 0;		/* if 1, make stdout and stderr explicitly 
+static int linebuffer = 0;	/* if 1, make stdout and stderr explicitly 
 				   line buffered, GNU libc makes them fully buffered
 				   if redirected to files */
 int timeout_long = 7;
 int timeout_short = 2;
 int timeout_active = 90;
-int timeout_client = 15*60;
+int timeout_client = 15*60;	/* when newsreader is idle for this many seconds, disconnect */
+int timeout_fetchnews = 5*60;	/* wait at most this many seconds for server replies in fetchnews */
 int clamp_maxage = 1;           /* if 1, maxage will be lowered to
 				 * groupexpire or expire if the
 				 * applicable parameter is lower than
@@ -77,47 +87,75 @@ int allowstrangers = 0;
 char *filterfile;
 struct server *servers = NULL;
 int allow_8bit_headers = 0;
+char *newsadmin;
+unsigned long timeout_lock = 5UL;
 
-static int parse_line( /*@unique@*/ char *l, /*@out@*/ char *param,	/*@out@*/
-		      char *value);
+/** parse the line in \a l, breaking it into param and value at the "="
+ * delimiter. The right-hand side can be quoted with double quotes,
+ * inside these a backslash escapes a quote that is part of the string.
+ * \return success
+ */
+static int parse_line(
+	/*@unique@*/ char *l /** input, will be modified */,
+	/*@out@*/ char *param /** output, left-hand side */,
+	/*@out@*/ char *value /** output, right-hand side */);
 
 /* parse a line, destructively */
-int
+static int
 parse_line(char *l, char *param, char *value)
 {
-    char *p, *q;
+    char *p;
+    size_t le, len;
+    enum modes { plain, quoted } mode = plain;
 
     p = l;
-    /* skip comments */
-    q = strchr(p, '#');
-    if (q)
-	*q = '\0';
-    if (*p == '\0')
-	return 0;
-    if ((q = strchr(p, '=')) == NULL)
-	return 0;
-    else
-	*q++ = '\0';
-    /* skipping leading blanks or tabs */
+    /* skip leading spaces, read parameter */
     SKIPLWS(p);
-    SKIPLWS(q);
-    (void)strlcpy(param, p, TOKENSIZE);
-    (void)strlcpy(value, q, TOKENSIZE);
+    le = strcspn(p, "=#");
+    /* strip trailing space */
+    while(le && strchr(" \t", p[le-1])) le--;
+    len = min(le, TOKENSIZE - 1);
+    if (!len) return 0;
+    memcpy(param, p, len);
+    param[len] = '\0';
+    p += le;
 
-    /* now param contains the stuff before '=' and value the stuff behind it */
-    if (!(*param) || !(*value))
+    SKIPLWS(p);
+    if (*p++ != '=')
 	return 0;
-    /* skipping trailing blanks or tabs */
-    p = param + strlen(param) - 1;
-    q = value + strlen(value) - 1;
-    while (isspace((unsigned char)*p) && (p > param)) {
-	*p = '\0';
-	p--;
+    SKIPLWS(p);
+
+    /* strip trailing blanks from input */
+    le = strlen(p);
+    while (le--) {
+	if (p[le] == ' ' || p[le] == '\t')
+	    p[le] = '\0';
+	else
+	    break;
     }
-    while (isspace((unsigned char)*q) && (q > value)) {
-	*q = '\0';
-	q--;
+
+    /* read value */
+    for (le = 0 ; le < TOKENSIZE - 1 ; le ++) {
+	char c = *p++;
+	if (mode == plain) {
+	    if (c == '#' || c == '\0') { break; }
+	    if (c == '"') { mode = quoted; continue; }
+	    *value++ = c;
+	} else if (mode == quoted) {
+	    if (c == '\\') {
+		if (*p) {
+		    *value++ = *p++; continue; 
+		} else
+		    return 0;
+	    }
+	    if (c == '\0') return 0;
+	    if (c == '"') break;
+	    *value++ = c;
+	} else {
+	    abort();
+	}
     }
+    *value = '\0';
     return 1;
 }
 
@@ -129,6 +167,37 @@ parse_line(char *l, char *param, char *value)
              Set p and q defaults to 0
 */
 
+/* parses value into timeout_lock, returns 0 for success, -1 for error */
+static int read_timeout_lock(
+	const char *value, /* input */
+	const char *source /* where did value come from */) {
+    char *t;
+    unsigned long u;
+
+    errno = 0;
+    u = strtoul(value, &t, 10);
+    if ((u != 0 || errno == 0)
+	    && t > value
+	    && (!*t || isspace((unsigned char)*t)))
+    {
+	timeout_lock = u;
+	if (debugmode) {
+	    if (timeout_lock) {
+		syslog(LOG_DEBUG,
+			"%s: waiting %lu seconds for lockfile",
+			source, timeout_lock);
+	    } else {
+		syslog(LOG_DEBUG,
+			"%s: waiting indefinitely for lockfile", source);
+	    }
+	}
+	return 0;
+    } else {
+	ln_log(LNLOG_SERR, LNLOG_CTOP, "error: cannot parse lockfile value \"%s\" from %s", value, source);
+	return -1;
+    }
+}
+
 int
 readconfig(int logtostderr)
 {
@@ -159,7 +228,7 @@ readconfig(int logtostderr)
 		if (p) {
 		    if (p->username != NULL)
 			free(p->username);
-		    p->username = strdup(value);
+		    p->username = critstrdup(value, "readconfig");
 		    if (debugmode)
 			syslog(LOG_DEBUG, "config: found username for %s",
 			       p->name);
@@ -169,7 +238,7 @@ readconfig(int logtostderr)
 		if (p) {
 		    if (p->password != NULL)
 			free(p->password);
-		    p->password = strdup(value);
+		    p->password = critstrdup(value, "readconfig");
 		    if (debugmode)
 			syslog(LOG_DEBUG, "config: found password for %s",
 			       p->name);
@@ -210,10 +279,14 @@ readconfig(int logtostderr)
 		expire = time(NULL) - (time_t) (SECONDS_PER_DAY * i);
 		if (debugmode)
 		    syslog(LOG_DEBUG, "config: expire is %d days", i);
+	    } else if (strcmp("newsadmin", param) == 0) {
+		if (debugmode)
+		    syslog(LOG_DEBUG, "config: newsadmin is %s", value);
+		newsadmin = critstrdup(value, "readconfig");
 	    } else if (strcmp("filterfile", param) == 0) {
 		if (debugmode)
 		    syslog(LOG_DEBUG, "config: filterfile is %s", value);
-		filterfile = strdup(value);
+		filterfile = critstrdup(value, "readconfig");
 	    } else if ((strcmp("hostname", param) == 0) ||
 		    (strcmp("fqdn", param) == 0)) {
 		if (is_validfqdn(value)) {
@@ -268,7 +341,9 @@ readconfig(int logtostderr)
 		    syslog(LOG_DEBUG, "config: allow_8bit_headers is %d",
 			    allow_8bit_headers);
 	    } else if (strcmp("debugmode", param) == 0) {
-		debugmode = atoi(value);
+		int d;
+		d = atoi(value);
+		debugmode = d > debugmode ? d : debugmode;
 		if (debugmode)
 		    syslog(LOG_DEBUG, "config: debugmode is %d", debugmode);
 	    } else if (strcmp("delaybody_in_situ", param) == 0) {
@@ -301,6 +376,13 @@ readconfig(int logtostderr)
 		if (debugmode)
 		    syslog(LOG_DEBUG, "config: timeout_client is %d secs",
 			   timeout_client);
+	    } else if (strcmp("timeout_fetchnews", param) == 0) {
+		timeout_fetchnews = atoi(value);
+		if (debugmode)
+		    syslog(LOG_DEBUG, "config: timeout_fetchnews is %d secs",
+			   timeout_fetchnews);
+	    } else if (strcmp("timeout_lock", param) == 0) {
+		read_timeout_lock(value, "config");
 	    } else if (strncmp("groupexpire", param, 11) == 0) {
 		char *m;
 		m = param;
@@ -343,7 +425,9 @@ readconfig(int logtostderr)
 		maxage = atoi(value);
 		if (maxage > LONG_MAX / 86400) {
 		    maxage = 24854; /* 32-bit: LONG_MAX / 86400 - 1 */
-		    ln_log(LNLOG_SWARNING, LNLOG_CTOP, "config: maxage clamped to %d", maxage);
+		    ln_log(LNLOG_SWARNING, LNLOG_CTOP,
+			    "warning: config: maxage cannot exceed %d, "
+			    "please fix %s", maxage, s);
 		}
 		if (debugmode)
 		    syslog(LOG_DEBUG, "config: maxage is %d", maxage);
@@ -376,6 +460,14 @@ readconfig(int logtostderr)
 		} else 
 		    syslog(LOG_ERR, "config: no server for noactive = %s", 
 			   value);
+	    } else if (strcmp("noxover", param) == 0) {
+		if (p) {
+		    p->noxover = TRUE;
+		    if (debugmode)
+			syslog(LOG_DEBUG, "config: no XOVER for %s",
+			       p->name);
+		} else
+		    syslog(LOG_ERR, "config: no server for noxover = %s", value);
 	    } else if (strcmp("nodesc", param) == 0) {
 		if (p) {
 		    p->descriptions = FALSE;
@@ -402,12 +494,20 @@ readconfig(int logtostderr)
 		    syslog(LOG_ERR, "config: no server for post_anygroup = %s", value);
 	    } else if (strcmp("noread", param) == 0) {
 		if (p) {
-    	    p->noread = atoi(value);
+		    p->noread = atoi(value);
 		    if (debugmode)
 			syslog(LOG_DEBUG, "config: noread for %s is %d",
 			       p->name, p->noread);
 		} else
 		    syslog(LOG_ERR, "config: no server for noread = %s", value);
+	    } else if (strcmp("only_groups_match_all", param) == 0) {
+		if (p) {
+		    p->only_groups_match_all = atoi(value);
+		    if (debugmode)
+			syslog(LOG_DEBUG, "config: only_groups_match_all for %s is %d",
+			       p->name, p->only_groups_match_all);
+		} else
+		    syslog(LOG_ERR, "config: no server for only_groups_match_all = %s", value);
 	    } else if (strcmp("initialfetch", param) == 0) {
 		initiallimit = strtoul(value, NULL, 10);
 		if (debugmode)
@@ -418,7 +518,7 @@ readconfig(int logtostderr)
 		if (debugmode)
 		    syslog(LOG_DEBUG, "config: server is %s", value);
 		p = (struct server *)critmalloc(sizeof(struct server),
-						    "allocating space for server name");
+						    "allocating space for server");
 		p->name = critstrdup(value, "readconfig");
 		p->descriptions = TRUE;
 		p->next = NULL;
@@ -428,9 +528,11 @@ readconfig(int logtostderr)
 		p->password = NULL;
 		p->nopost = 0;
 		p->noread = 0;
+		p->noxover = 0;
 		p->post_anygroup = 0;
 		p->updateactive = TRUE;
 		p->group_pcre = NULL;
+		p->only_groups_match_all = 0;
 		if (servers == NULL)
 		    servers = p;
 		else
@@ -445,6 +547,7 @@ readconfig(int logtostderr)
 			syslog(LOG_DEBUG, "config: only_groups_pcre for %s is %s",
 			       p->name, value);
 		} else {
+		    free(re);
 		    syslog(LOG_ERR, "config: no server for nopost = %s", value);
 		}
 	    } else {
@@ -469,18 +572,53 @@ readconfig(int logtostderr)
     }
     debug = debugmode;
 
+    if (!newsadmin) {
+	const char t[] = NEWS_USER;
+	newsadmin = critmalloc(strlen(fqdn) + strlen(t) + 2, "readconfig");
+	strcpy(newsadmin, t); /* RATS: ignore */
+	strcat(newsadmin, "@");
+	strcat(newsadmin, fqdn); /* RATS: ignore */
+    }
+
     expire_base = ent;
     fclose(f);
+    free(param);
+    free(value);
+
     if (servers == NULL) {
 	syslog(LOG_ERR, "no server declaration in config file");
-	free(param);
-	free(value);
 	return 0;
     }
     if (!expire)
 	syslog(LOG_ERR, "no expire declaration in config file");
 
+    /* check for duplicate server configurations */
+    {
+	unsigned short port = 0;
+	struct servent *sp = getservbyname("nntp", "tcp");
+	if (sp) port = ntohs(sp->s_port);
+
+	for (p = servers; p ; p=p->next) {
+	    for (q = p->next ; q ; q=q->next) {
+		unsigned short pp = p->port, qp = q->port;
+		if (!pp) pp = port;
+		if (!qp) qp = port;
+		if (!pp || !qp) {
+		    syslog(LOG_ERR, "Cannot resolve service \"nntp\" protocol \"tcp\".");
+		    fprintf(stderr, "Cannot resolve service \"nntp\" protocol \"tcp\".\n");
+		    return 0;
+		}
+		if (pp != qp) continue;
+		if (strcasecmp(p->name, q->name)) continue;
+		syslog(LOG_ERR, "Duplicate definition for server %s port %hu", p->name, pp);
+		fprintf(stderr, "Duplicate definition for server %s port %hu\n", p->name, pp);
+		return 0;
+	    }
+	}
+    }
+
     if (debugmode > 1) {
+	getrlimit(RLIMIT_CORE, &corelimit);
 	corelimit.rlim_cur = COREFILESIZE;
 	if (setrlimit(RLIMIT_CORE, &corelimit) < 0)
 	    syslog(LOG_DEBUG, "Changing core file size failed: %m");
@@ -489,6 +627,10 @@ readconfig(int logtostderr)
 	syslog(LOG_DEBUG, "Core file size: %d", (int)corelimit.rlim_cur);
     }
 
+    l = getenv("LN_LOCK_TIMEOUT");
+    if (l && *l)
+	read_timeout_lock(l, "LN_LOCK_TIMEOUT");
+
     if (linebuffer) {
 	fflush(stdout);
 	setvbuf(stdout, NULL, _IOLBF, BUFSIZ);
@@ -496,8 +638,6 @@ readconfig(int logtostderr)
 	setvbuf(stderr, NULL, _IOLBF, BUFSIZ);
     }
 
-    free(param);
-    free(value);
     validatefqdn(logtostderr);
     return 1;
 }
@@ -522,12 +662,14 @@ lookup_expireent(char *group)
 
 static void
 freeserver(struct server *a) {
+    if (a->group_pcre) pcre_free(a->group_pcre);
     if (a->name) free(a->name);
     if (a->username) free(a->username);
     if (a->password) free(a->password);
+    free(a);
 }
 
-void
+void /* exported for exclusive use in nntpd.c */
 freeservers(void) {
     struct server *i = servers, *n;
 
@@ -536,6 +678,36 @@ freeservers(void) {
 	freeserver(i);
 	i = n;
     }
+    servers = NULL;
+}
+
+
+void
+freeexpire(void)
+{
+    struct expire_entry *a, *b;
+
+    a = expire_base;
+    while(a)
+    {
+	b = a->next;
+	free(a->group);
+	free(a);
+	a = b;
+    }
+}
+
+void freeconfig(void) {
+    freeservers();
+    if (newsadmin)
+	free(newsadmin);
+    freefilter();
+    if (filterfile)
+	free(filterfile);
+    freegetaline();
+    freeexpire();
+    (void)lookup(LOOKUP_FREE);
+    freelastreply();
 }
 
 time_t lookup_expire(char *group)
