@@ -3,7 +3,7 @@
  * conditions.
  */
 
-/* $Id: leafnode.h,v 1.61 2003/10/14 18:39:16 emma Exp $ */
+/* $Id: leafnode.h,v 1.91 2005/03/06 19:20:57 emma Exp $ */
 
 #ifndef LEAFNODE_H
 #define LEAFNODE_H
@@ -18,18 +18,21 @@
 #define FALSE (0)
 #endif
 
-/* limits.h is supposed to contain PATH_MAX, we include sys/param.h too */
+/* limits.h may contain PATH_MAX but shall not if the system has variable
+ * length limits here */
 #include <limits.h>
 #ifndef PATH_MAX
-#include <sys/param.h>
-#define PATH_MAX MAXPATHLEN
+/* just some random limit - we'll use this as a starting point
+ * for dynamically growing memory only */
+#define PATH_MAX 4096
 #endif
 
 #define PORTFILENAMECSET "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789._-"
 
-#include <pcre.h>		/* pcre */
+#include <pcre.h>
 #include <stdarg.h>		/* va_list */
 #include <stdio.h>		/* FILE * */
+#include <setjmp.h>		/* sigjmp_buf */
 #include "system.h"		/* time.h */
 
 #include "config.h"		/* FreeSGI barfs on #ifdef HAVE_CONFIG_H */
@@ -53,17 +56,28 @@ char *strdup(const char *);
 
 /* from snprintf.c */
 #ifndef HAVE_WORKING_SNPRINTF
-int snprintf(char *str, size_t n, const char *format, ...);
-int vsnprintf(char *str, size_t n, const char *format, va_list ap);
+int ln_snprintf(char *str, size_t n, const char *format, ...)
+#ifdef __GNUC__
+    __attribute__ ((format(printf, 3, 4)))
+#endif
+    ;
+int ln_vsnprintf(char *str, size_t n, const char *format, va_list ap);
+#define snprintf ln_snprintf
+#define vsnprintf ln_vsnprintf
 #endif
 
-int xsnprintf(/*@out@*/ char *str, size_t n, const char *format, ...);
+int xsnprintf(/*@out@*/ char *str, size_t n, const char *format, ...)
+#ifdef __GNUC__
+    __attribute__ ((format(printf, 3, 4)))
+#endif
+    ;
 
-/* add LOG_NEWS where it doesn't exist */
+/* map LOG_NEWS onto LOG_DAEMON where the former doesn't exist */
 #include <syslog.h>
 #if !defined( LOG_NEWS )
 #define LOG_NEWS LOG_DAEMON
 #endif
+/* define LOG_CONS if missing */
 #if !defined( LOG_CONS )
 #define LOG_CONS 0		/* if it isn't supported, make do without */
 #endif
@@ -78,16 +92,16 @@ int initvars(char *progname);
 
 /* converts a message-id to a file name, the return value points into
    a static array */
+#define LOOKUP_FREE ((const char *)-2)
 /*@dependent@*/ const char *lookup(const char *msgid);
 
 /*@dependent@*/
 char *getaline(FILE * f);	/* reads one line, regardless of length */
 /*@dependent@*/
 char *mgetaline(FILE * f);      /* dito, with timeout */
-/*@dependent@*/
-char *tgetaline(FILE * f);	/* read line with timer */
-
-RETSIGTYPE timer(int sig);
+void mgetaline_settimeout(unsigned int);      /* set timeout for mgetaline */
+extern sigjmp_buf timeout;
+extern RETSIGTYPE timer(int sig);
 
 /* changes (and optionally creates) directory */
 int chdirgroup(const char *group, int creatdir);
@@ -106,7 +120,8 @@ struct newsgroup {
 int isinteresting(const char *groupname);
 void insertgroup(const char *name, long unsigned first, long unsigned last,
 		 time_t date);
-void changegroupdesc(const char *groupname, char *description);
+extern void changegroupdesc(const char *groupname, const char *description);
+extern void newgroupdesc(const char *groupname, const char *description);
 void mergegroups(void);
 /*@null@*//*@dependent@*/
 struct newsgroup *xfindgroup(struct newsgroup *active, const char *name,
@@ -116,7 +131,9 @@ struct newsgroup *findgroup(const char *name);
 void readactive(void);
 int writeactive(void);
 void fakeactive(void);
-void freeactive(struct newsgroup *act);
+void freeactive(/*@null@*/ struct newsgroup *act);
+extern char *activeread(void);
+extern int killactiveread(void);
 
 extern size_t activesize;
 /*@null@*/ extern struct newsgroup *active;
@@ -126,8 +143,10 @@ extern size_t oldactivesize;
 /* translation from message-id to article number, used in fetch and expire */
 
 void clearidtree(void);
-void insertmsgid( /*@unique@*/ const char *msgid, unsigned long art);
-unsigned long findmsgid(const char *msgid);
+void insertmsgid( /*@unique@*/ const char *msgid);
+int findmsgid(const char *msgid);
+typedef int (*tmihook)(const char *);
+int traverseidtree(tmihook h);
 
 /* -----------here starts the new stuff-----------------*/
 
@@ -143,7 +162,7 @@ void
 prependtolist(struct stringlist **list, /*@unique@*/ const char *newentry);
 
 /*@null@*//*@dependent@*/
-struct stringlist **lfindinlist(struct stringlist **haystack, char *needle, int len);
+struct stringlist **lfindinlist(struct stringlist **haystack, char *needle, size_t len);
 	/* find a stringlist element by doing a linear search */
 char *findinlist(struct stringlist *haystack, char *needle);
 	/* find a string in a stringlist by doing a linear search */
@@ -151,17 +170,14 @@ void freelist( /*@only@*/ struct stringlist *list);
 	/* free memory occupied by a stringlist */
 
 /*
- * filtering headers for PCRE
+ * filterfile.c -- PCRE filtering of articles
  */
 void readfilter(char *filterfile);
+void freefilter(void);
 int dofilter(char *h);
 
 /*
- * artutil -- handling article files
- */
-
-/*
- * store articles
+ * artutil.c -- handling article files
  */
 void store(const char *filename,
 	   FILE * filehandle, const char *newsgroups, const char *msgid);
@@ -172,20 +188,6 @@ void store(const char *filename,
 /*@null@*//*@only@*/ char *getheader(const char *filename, const char *header);
 /*@null@*//*@only@*/ char *fgetheader(FILE * f, const char *header);
 
-/*
- * xover stuff -- for nntpd.c
- */
-struct xoverinfo {
-    char *text;
-    int exists;
-};
-
-/*@null@*/ extern struct xoverinfo *xoverinfo;
-extern unsigned long xfirst;
-extern unsigned long xlast;
-
-int getxover(void);		/* set xoverinfo, return 0 on error, nonzero else */
-
 /*
  * the strings in config.c
  */
@@ -219,6 +221,9 @@ struct server {
     int updateactive;		/* update the active file of this server */
     int post_anygroup;		/* if set, do not check if the group is on the
 				   server, but just post */
+    int noxover;		/* if set, use XHDR for overview */
+    int only_groups_match_all;	/* if set, any unmatched group (refer to
+				   group_pcre) prevents the post */
 };
 
 extern int date_is_evil;	/* skip DATE check on servers known bad */
@@ -248,13 +253,16 @@ extern int timeout_long;	/* don't fetch groups that have been accessed
 				   that many days */
 extern int timeout_active;	/* reread active file after that many days */
 extern int timeout_client;	/* activity timeout for clients in seconds */
-extern int clamp_maxage;        /* limit maxage to applicable group/global expire? */
-extern int article_despite_filter;   /* request and discard body if filterfile is defined, for high-latency, high-throughput links */
+extern int timeout_fetchnews;	/* response deadline for upstream in seconds */
+extern int clamp_maxage;	/* limit maxage to applicable group/global expire? */
+extern int article_despite_filter;  /* request and discard body if filterfile is defined, for high-latency, high-throughput links */
 extern char *filterfile;	/* filename where filter resides */
 extern struct server *servers;	/* list of servers to use */
 extern int allowstrangers;	/* if addresses not local to our links
 				   are allowed to connect */
 extern int allow_8bit_headers;	/* if 8bit junk in headers is allowed */
+extern char *newsadmin;		/* address of news administrator, default: news@%s with %s = fqdn */
+extern unsigned long timeout_lock; /* lock file timeout */
 
 /*
  * other global variables
@@ -264,14 +272,8 @@ extern char last_command[SIZE_lineout + 1];
 extern char lineout[SIZE_lineout + 1];
 
 /* defined in nntputil.c */
-extern
-				    /*@relnull@*/
- /*@dependent@*/
-FILE *nntpin;
-extern
-				    /*@relnull@*/
- /*@dependent@*/
-FILE *nntpout;
+extern /*@relnull@*/ /*@dependent@*/ FILE *nntpin;
+extern /*@relnull@*/ /*@dependent@*/ FILE *nntpout;
 
 #define SIZE_s (8192)
 #define FQDNLEN 255
@@ -288,13 +290,14 @@ int handover_lock(pid_t);
 void putaline(void);
 void retry(void);
 void readexpire(void);
-void free_expire(void);
 int readconfig(int logtostderr);
+void freeexpire(void);
 void freeservers(void);
+void freeconfig(void);
 void lowercase(char *string);
 int ngmatch(const char *pattern, const char *string);
 void overrun(void);
-void replaceinlist(struct stringlist **haystack, char *needle, int len);
+void replaceinlist(struct stringlist **haystack, char *needle, size_t len);
 time_t lookup_expire(char *group);
 int lookup_expiredays(char *group);
 int log_unlink(const char *f, int ignore_enoent);
@@ -312,6 +315,8 @@ int nntpreply(const struct server*);	/* decode an NNTP reply number */
 int nntpconnect(const struct server *upstream);
 				/* connect to upstream server */
 void nntpdisconnect(void);	/* disconnect from upstream server */
+void nntpquit(void);		/* send QUIT, then disconnect from upstream server */
+void freelastreply(void);
 
 /*@dependent@*/
 const char *rfctime(void);	/* An rfc type date */
@@ -329,9 +334,11 @@ int mkstemp(char *);
 #ifndef HAVE_GETLINE
 ssize_t getline(char **, size_t *, FILE *);	/* fgets replacement */
 #endif
+void freegetaline(void);
 
-/* from gmtoff.c */
-time_t gmtoff(const time_t);
+#ifndef HAVE_TIMEGM
+time_t timegm(struct tm *tm);
+#endif
 
 /* from mysetvbuf.c */
 int mysetvbuf(FILE *stream, char *buf, int mode, size_t size);
@@ -363,6 +370,24 @@ int pcre_extract(const char *input, const char *pattern, /*@out@*/ char **output
 void pcre_extract_free(char **vec, int count);
 size_t xstrlcpy(/*@out@*/ char *dst, const char *src, size_t size);
 
+/* xoverutil.c */
+struct xoverinfo {
+    char *text;
+    int exists;
+};
+
+/*@null@*/ extern struct xoverinfo *xoverinfo;
+extern unsigned long xfirst;
+extern unsigned long xlast;
+
+int getxover(void);		/* set xoverinfo, return 0 on error, nonzero else */
+
+int legalxoverline(const char *xover, /*@out@*/ const char **errmsg);
+void freexover(void);
+
+/* agetcwd.c */
+int agetcwd(/*@out@*/ char **buf, /*@out@*/ size_t *capa);
+
 #ifdef CHECKGROUPORDER
 void checkgrouporder(void);
 #endif /* CHECKGROUPORDER */
