@@ -70,41 +70,15 @@ union sockaddr_union {
 };
 #endif
 
-int hash(const char *);
-/*@null@*/ /*@dependent@*/ static FILE *fopenart(const char *);
-/*@null@*/ /*@dependent@*/ static FILE *fopenpseudoart(const char *arg, const unsigned long article_num);
-/*@null@*/ /*@dependent@*/ static FILE *buildpseudoart(const char *grp);
-void list(struct newsgroup *ng, int what, const char *pattern);
-void rereadactive(void);
-
-void parser(void);
-void myerror(const char *msg);
-void doarticle(const char *arg, int what);
-int dogroup(const char *);
-void dohelp(void);
-void domove(int);
-void dolist(char *p);
-void donewgroups(const char *);
-/* void donewnews(const char*); */
-int dopost(void);
-void doxhdr(char *);
-void doxover(char *);
-void dolistgroup(const char *);
-static void markinterest(const char *);
-static void nogroup(void);
-
 /*@null@*/ static struct newsgroup *group;	/* current group, initially none */
 /*@null@*/ static struct newsgroup *xovergroup = NULL;
-unsigned long artno;		/* current article number */
-/*@dependent@*/ /*@null@*/ char *cmd;	/* current command line */
-time_t activetime;
+static unsigned long artno;		/* current article number */
+/*@dependent@*/ /*@null@*/ static char *cmd;	/* current command line */
+static time_t activetime;
 
 int debug = 0;
 int verbose = 0;		/* verbose doesn't count here */
 
-static int parserange(char *arg, unsigned long *a, unsigned long *b);
-static int is_pseudogroup(/*@null@*/ const struct newsgroup *g);
-static void fatal_write(void);
 static void
 fatal_write(void)
 {
@@ -114,7 +88,7 @@ fatal_write(void)
     exit(EXIT_FAILURE);
 }
 
-void
+static void
 rereadactive(void)
 {
     struct stat st;
@@ -142,126 +116,6 @@ rereadactive(void)
     }
 }
 
-
-
-void
-myerror(const char *msg)
-{
-    printf("%s: %s\r\n", msg, strerror(errno));
-}
-
-
-void
-parser(void)
-{
-    char *arg;
-    int n;
-    size_t size;
-
-    while ((cmd = mgetaline(stdin))) {
-	if (debug == 1)
-	    syslog(LOG_DEBUG, "<%s", cmd);
-
-	size = strlen(cmd);
-	if (size == 0)
-	    continue;		/* ignore */
-	if (size > MAXLINELENGTH || (long)size > (long)INT_MAX) {
-	    /* ignore attempts at buffer overflow */
-	    if (debugmode)
-		syslog(LOG_DEBUG, ">500 Dazed and confused");
-	    printf("500 Dazed and confused\r\n");
-	    continue;
-	}
-
-	/* parse command line */
-	n = 0;
-	while (isalpha((unsigned char)cmd[n]))
-	    n++;
-	while (isspace((unsigned char)cmd[n]))
-	    cmd[n++] = '\0';
-
-	arg = cmd + n;
-
-	while (cmd[n])
-	    n++;
-	n--;
-	while (n >= 0 && isspace((unsigned char)cmd[n]))
-	    cmd[n--] = '\0';
-
-	if (!strcasecmp(cmd, "quit")) {
-	    if (debugmode)
-		syslog(LOG_DEBUG, ">205 Always happy to serve!");
-	    printf("205 Always happy to serve!\r\n");
-	    return;
-	}
-	rereadactive();
-	if (!strcasecmp(cmd, "article")) {
-	    doarticle(arg, 3);
-	} else if (!strcasecmp(cmd, "head")) {
-	    doarticle(arg, 2);
-	} else if (!strcasecmp(cmd, "body")) {
-	    doarticle(arg, 1);
-	} else if (!strcasecmp(cmd, "stat")) {
-	    doarticle(arg, 0);
-	} else if (!strcasecmp(cmd, "help")) {
-	    dohelp();
-	} else if (!strcasecmp(cmd, "last")) {
-	    domove(-1);
-	} else if (!strcasecmp(cmd, "next")) {
-	    domove(1);
-	} else if (!strcasecmp(cmd, "list")) {
-	    dolist(arg);
-	} else if (!strcasecmp(cmd, "date")) {
-	    dodate();
-	} else if (!strcasecmp(cmd, "mode")) {
-	    if (debugmode)
-		syslog(LOG_DEBUG, ">200 Leafnode %s, pleased to meet you!",
-		       version);
-	    printf("200 Leafnode %s, pleased to meet you!\r\n", version);
-	} else if (!strcasecmp(cmd, "newgroups")) {
-	    donewgroups(arg);
-	} else if (!strcasecmp(cmd, "newnews")) {
-	    if (debugmode)
-		syslog(LOG_DEBUG,
-		       ">500 NEWNEWS is meaningless for this server");
-	    printf("500 NEWNEWS is meaningless for this server\r\n");
-	} else if (!strcasecmp(cmd, "post")) {
-	    if (dopost()) break;
-	} else if (!strcasecmp(cmd, "slave")) {
-	    if (debugmode)
-		syslog(LOG_DEBUG, ">202 Cool - I always wanted a slave");
-	    printf("202 Cool - I always wanted a slave\r\n");
-	} else if (!strcasecmp(cmd, "xhdr")) {
-	    doxhdr(arg);
-	} else if (!strcasecmp(cmd, "hdr")) {
-	    doxhdr(arg);
-	} else if (!strcasecmp(cmd, "xover")) {
-	    doxover(arg);
-	} else if (!strcasecmp(cmd, "over")) {
-	    doxover(arg);
-	} else if (!strcasecmp(cmd, "listgroup")) {
-	    dolistgroup(arg);
-	} else if (!strcasecmp(cmd, "group")) {
-	    if (dogroup(arg)) break;
-	} else {
-	    if (debugmode)
-		syslog(LOG_DEBUG, ">500 Unknown command");
-	    printf("500 Unknown command\r\n");
-	}
-	if (ferror(stdout) || fflush(stdout)) {
-	    syslog(LOG_ERR, "Cannot write to client.");
-	    break;
-	}
-    }
-    if (debugmode)
-	syslog(LOG_DEBUG, "Client timeout, disconnecting.");
-
-    /* There was once a 400 error message here. It confused broken
-     * clients, most notably, tin.
-     * Future NNTP drafts command that we don't send stuff back on
-     * timeout, so we anticipate these. */
-}
-
 /*
  * pseudo article stuff
  */
@@ -294,7 +148,7 @@ buildpseudoart(const char *grp)
 
     fprintf(f, "Path: %s\n", fqdn);
     fprintf(f, "Newsgroups: %s\n", grp);
-    fprintf(f, "From: Leafnode <news@%s>\n", fqdn);
+    fprintf(f, "From: Leafnode <%s>\n", newsadmin);
     fprintf(f, "Subject: Leafnode placeholder for group %s\n", grp);
     fprintf(f, "Date: %s\n", rfctime());
     fprintf(f, "Message-ID: <leafnode:placeholder:%s@%s>\n", grp, fqdn);
@@ -306,20 +160,20 @@ buildpseudoart(const char *grp)
 	    "\n"
 	    "If you do an operation on a group - such as reading an article,\n"
 	    "looking at the group table of contents or similar, then leafnode\n"
-	    "will go and fetch articles from that group when it next updates.\n"
-	    "\n"
+	    "will go and fetch articles from that group when it next updates.\n\n");
+    fprintf(f,
 	    "Since you have read this dummy article, leafnode will retrieve\n"
 	    "the newsgroup %s when fetchnews is run\n"
 	    "the next time. If you'll look into this group a little later, you\n"
-	    "will see real articles.\n"
-	    "\n"
+	    "will see real articles.\n\n", grp);
+    fprintf(f,
 	    "If you see articles in groups you do not read, that is almost\n"
 	    "always because of cross-posting.  These articles do not occupy any\n"
 	    "more space - they are hard-linked into each newsgroup directory.\n"
 	    "\n"
 	    "If you do not understand this, please talk to your newsmaster.\n"
 	    "\n"
-	    "Leafnode can be found at\n" "\thttp://www.leafnode.org/\n\n", grp);
+	    "Leafnode can be found at\n" "\thttp://www.leafnode.org/\n\n");
 
     if (ferror(f)) {
 	(void)fclose(f);
@@ -331,6 +185,89 @@ buildpseudoart(const char *grp)
     return f;
 }
 
+static int
+parserange(char *arg, unsigned long *a, unsigned long *b)
+{
+    char *l = arg;
+    /* no argument */
+    if (!*l) return 0;
+
+    /* parse */
+    if (*l != '-')
+	*a = strtoul(arg, &l, 10);
+    SKIPLWS(l);
+    if (*l == '-') {
+	++l;
+	SKIPLWS(l);
+	if (isdigit((unsigned char)*l))
+	    *b = strtoul(l, &l, 10);
+    } else {
+	*b = *a;
+    }
+    SKIPLWS(l);
+
+    /* trailing garbage */
+    if (l && *l) {
+	return 0;
+    }
+
+    return 1;
+}
+
+static int
+is_pseudogroup(/*@null@*/ const struct newsgroup *g)
+{
+    if (!g) return 0;
+    if (!chdirgroup(g->name, FALSE)) return 1;
+/*    if (isinteresting(g->name)) return 0; */
+    if (g->last < g->first) return 1;
+    if (g->last <= 1) return 1;
+    return 0;
+}
+
+static void
+markinterest(const char *ng)
+{
+    struct stat st;
+    struct utimbuf buf;
+    int err;
+    time_t now;
+    FILE *f;
+    char s[SIZE_s+1];
+
+    err = 0;
+    (void)xsnprintf(s, SIZE_s, "%s/interesting.groups/%s", spooldir, ng);
+    if (stat(s, &st) == 0) {
+	now = time(NULL);
+	buf.actime = (now < st.st_atime) ? st.st_atime : now;
+	/* now < update may happen through HW failures */
+	buf.modtime = st.st_mtime;
+	if (utime(s, &buf)) {
+	    err = 1;
+	}
+	if (debugmode)
+	    syslog(LOG_DEBUG, "markinterest: %s touched", ng);
+    } else {
+	err = 1;
+    }
+    if (err) {
+	f = fopen(s, "w");
+	if (f == NULL || fclose(f) == EOF) {
+	    syslog(LOG_ERR, "Could not create %s: %m", s);
+	} else {
+	    if (debugmode)
+		syslog(LOG_DEBUG, "markinterest: %s new", ng);
+	}
+    } else {
+	int fd = open(".", O_RDONLY);
+	(void)chdirgroup(ng, FALSE);
+	if (fd >= 0) {
+	    (void)fchdir(fd);
+	    (void)close(fd);
+	}
+    }
+}
+
 /* open a pseudo art */
 /* WARNING: article_num MUST be 0 for selection based on Message-ID */
 static FILE *
@@ -364,9 +301,8 @@ fopenpseudoart(const char *arg, const unsigned long article_num)
     return f;
 }
 
-
 /* open an article by number or message-id */
-FILE *
+static FILE *
 fopenart(const char *arg)
 {
     unsigned long int a;
@@ -394,16 +330,16 @@ fopenart(const char *arg)
 	} else {
 	    f = fopen(ptr, "r");
 	}
-	if (f != NULL && (fstat(fileno(f), &st) || st.st_size == 0)) {
-	    (void)fclose(f);
-	    f = NULL;
-	}
 	if (f != NULL)
 	    artno = a;
 	markinterest(group->name);
     } else {
 	f = NULL;
     }
+    if (f != NULL && (fstat(fileno(f), &st) || st.st_size == 0)) {
+	(void)fclose(f);
+	f = NULL;
+    }
     return f;
 }
 
@@ -452,9 +388,18 @@ markdownload(const char *ng, unsigned long id)
     return 1;
 }
 
+static void
+nogroup(void)
+{
+    printf("412 Use the GROUP command first\r\n");
+    if (debugmode)
+	syslog(LOG_DEBUG, ">412 Use the GROUP command first");
+    return;
+}
+
 /* display an article or somesuch */
 /* DOARTICLE */
-void
+static void
 doarticle(const char *arg, int what)
 {
     FILE *f;
@@ -664,51 +609,8 @@ doarticle(const char *arg, int what)
 }
 
 
-static void
-markinterest(const char *ng)
-{
-    struct stat st;
-    struct utimbuf buf;
-    int err;
-    time_t now;
-    FILE *f;
-    char s[SIZE_s+1];
-
-    err = 0;
-    (void)xsnprintf(s, SIZE_s, "%s/interesting.groups/%s", spooldir, ng);
-    if (stat(s, &st) == 0) {
-	now = time(NULL);
-	buf.actime = (now < st.st_atime) ? st.st_atime : now;
-	/* now < update may happen through HW failures */
-	buf.modtime = st.st_mtime;
-	if (utime(s, &buf)) {
-	    err = 1;
-	}
-	if (debugmode)
-	    syslog(LOG_DEBUG, "markinterest: %s touched", ng);
-    } else {
-	err = 1;
-    }
-    if (err) {
-	f = fopen(s, "w");
-	if (f == NULL || fclose(f) == EOF) {
-	    syslog(LOG_ERR, "Could not create %s: %m", s);
-	} else {
-	    if (debugmode)
-		syslog(LOG_DEBUG, "markinterest: %s new", ng);
-	}
-    } else {
-	int fd = open(".", O_RDONLY);
-	(void)chdirgroup(ng, FALSE);
-	if (fd >= 0) {
-	    (void)fchdir(fd);
-	    (void)close(fd);
-	}
-    }
-}
-
 /* change to group - note no checks on group name */
-int
+static int
 dogroup(const char *arg)
 {
     struct newsgroup *g;
@@ -716,7 +618,13 @@ dogroup(const char *arg)
     g = findgroup(arg);
     if (g) {
 	group = g; /* global */
+	if (isinteresting(arg)) {
+	    if (debugmode)
+		syslog(LOG_DEBUG, "marked group %s interesting", arg);
+	    markinterest(arg);
+	}
 	if (!is_pseudogroup(g)) {
+	    /* regular news group */
 	    if (debugmode)
 		syslog(LOG_DEBUG, ">211 %lu %lu %lu %s group selected",
 		       g->last >= g->first ? g->last - g->first + 1 : 0,
@@ -725,7 +633,7 @@ dogroup(const char *arg)
 		   g->last >= g->first ? g->last - g->first + 1 : 0,
 		   g->first, g->last, g->name);
 	} else {
-	    g->first = g->last;
+	    /* pseudo news group */
 	    if (debugmode)
 		syslog(LOG_DEBUG,
 		       ">211 %lu %lu %lu %s group selected (pseudo article)",
@@ -743,7 +651,7 @@ dogroup(const char *arg)
     return 0;
 }
 
-void
+static void
 dohelp(void)
 {
     fputs("100 Legal commands on THIS server:\r\n"
@@ -770,7 +678,7 @@ dohelp(void)
 	    ".\r\n", stdout);
 }
 
-void
+static void
 domove(int by)
 {
     char *msgid;
@@ -819,36 +727,62 @@ domove(int by)
     }
 }
 
-
+static int is_pattern(const char *s) {
+    return s ? strcspn(s, "?*[") < strlen(s) : 1;
+}
 
 /* LIST ACTIVE if what==0,
  * LIST NEWSGROUPS if what==1
  * LIST ACTIVE.TIMES if what==2
  */
-void
-list(struct newsgroup *g, int what, const char *pattern)
+static void
+printlist(const struct newsgroup *g, const int what)
 {
-    while (g->name) {
-	if (!pattern || !ngmatch(pattern, g->name)) {
-	    switch(what) {
-		case 0:
-		    printf("%s %lu %lu y\r\n", g->name, g->last, g->first);
-		    break;
-		case 1:
-		    printf("%s\t%s\r\n", g->name, g->desc ? g->desc : "-x-");
-		    break;
-		case 2:
-		    printf("%s %lu news@%s\r\n", g->name, (unsigned long)g->age, fqdn);
-		    break;
-		default:
-		    break;
-	    }
-	}
-	g++;
+    switch(what) {
+	case 0:
+	    if (is_pseudogroup(g))
+		printf("%s %lu %lu y\r\n", g->name, g->first, g->first);
+	    else
+		printf("%s %lu %lu y\r\n", g->name, g->last, g->first);
+	    break;
+	case 1:
+	    printf("%s\t%s\r\n", g->name, g->desc ? g->desc : "-x-");
+	    break;
+	case 2:
+	    printf("%s %lu %s\r\n", g->name, (unsigned long)g->age, newsadmin);
+	    break;
+	default:
+	    abort();
+	    break;
     }
 }
 
-void
+/* LIST ACTIVE if what==0,
+ * LIST NEWSGROUPS if what==1
+ * LIST ACTIVE.TIMES if what==2
+ */
+static void
+list(struct newsgroup *g, const int what, const char *pattern)
+{
+    if (is_pattern(pattern)) {
+	while(g->name) {
+	    if (!pattern || !ngmatch(pattern, g->name)) {
+		printlist(g, what);
+	    }
+	    g++;
+	}
+    } else {
+	/* single group */
+	g = findgroup(pattern);
+	if (g) {
+	    printlist(g, what);
+	    if (what == 0 && isinteresting(pattern))
+		markinterest(pattern);
+	}
+    }
+}
+
+static void
 dolist(char *arg)
 {
     if (!strcasecmp(arg, "extensions")) {
@@ -874,7 +808,7 @@ dolist(char *arg)
 		".\r\n");
 	if (debugmode)
 	    syslog(LOG_DEBUG, ">215 information follows");
-    } else if (!strcasecmp(arg, "active.times")) {
+    } else if (!strncasecmp(arg, "active.times", 12)) {
 	if (active) {
 	    const char m[] = "215 Note that leafnode will fetch groups on demand.";
 		printf("%s\r\n", m);
@@ -934,7 +868,7 @@ dolist(char *arg)
     }
 }
 
-void
+static void
 donewgroups(const char *arg)
 {
     struct tm timearray;
@@ -990,14 +924,10 @@ donewgroups(const char *arg)
     timearray.tm_sec = (int)(a % 100);
     /* mktime() shall guess correct value of tm_isdst (0 or 1) */
     timearray.tm_isdst = -1;
-    /* mktime() assumes local time -> correct according to timezone
-       (global variable set by last call to localtime() ) if GMT is not
-       requested */
-    age = mktime(&timearray);
-    if (age != (time_t) - 1 && 0 == strncasecmp(l, "gmt", 3)) {
-	time_t off = gmtoff(age);
-	age += off;
-    }
+    if (0 == strncasecmp(l, "gmt", 3))
+	age = timegm(&timearray);
+    else
+	age = mktime(&timearray);
 
     printf("231 List of new newsgroups since %ld follows\r\n", (long)age);
     if (debugmode)
@@ -1065,7 +995,7 @@ generateMessageID(void)
 /* the end of what I stole from rsalz and then mangled */
 
 
-int
+static int
 dopost(void)
 {
     char *line;
@@ -1085,6 +1015,11 @@ dopost(void)
     char *suggmid;
     /*@observer@*/ const char *appendheader = NULL;
 
+    if (getenv("LN_REJECT_POST_PRE")) {
+	printf("400 Posting rejected - debug variable LN_REJECT_POST_PRE exists\r\n");
+	return 0;
+    }
+
     do {
 	(void)xsnprintf(outname, sizeof(outname), "%s/out.going/%d-%d-%d",
 		spooldir, (int)getpid(), (int)time(NULL), ++postingno);
@@ -1227,7 +1162,7 @@ dopost(void)
 		    err = 1;
 		if (fputs(fqdn, out) == EOF)
 		    err = 1;
-		if (fputs("!news\r\n", out) == EOF)
+		if (fprintf(out, "!%s\r\n", NEWS_USER) < 0)
 		    err = 1;
 	    }
 	    if (!havedate) {
@@ -1299,6 +1234,9 @@ dopost(void)
     if (!havefrom)
 	ferrstr = "Missing From: header";
 
+    if (getenv("LN_REJECT_POST_POST"))
+	ferr = 1;
+
     if (havefrom && havesubject && havenewsgroups && !ferr) {
 	if (!err && 0 == chmod(outname, 0644)) {
 	    printf("240 Article posted, now be patient\r\n");
@@ -1315,6 +1253,12 @@ dopost(void)
 
     (void)unlink(outname);
 
+    if (getenv("LN_REJECT_POST_POST")) {
+	printf("400 Posting rejected - debug variable LN_REJECT_POST_POST exists\r\n");
+	syslog(LOG_INFO, ">400 Posting rejected - debug variable LN_REJECT_POST_POST exists\r\n");
+	return 0;
+    }
+
     if (ferrstr) {
 	printf("441 Post rejected, formatting error: %s\r\n", ferrstr);
 	syslog(LOG_INFO, ">441 Post rejected, formatting error: %s", ferrstr);
@@ -1352,7 +1296,7 @@ static int checkrange(const struct newsgroup *g,
 
 
 
-void
+static void
 doxhdr(char *arg)
 {
     static const char *h[] = { "Subject", "From", "Date", "Message-ID",
@@ -1469,7 +1413,7 @@ doxhdr(char *arg)
 	    if (n == 0)		/* Subject: */
 		printf("Leafnode placeholder for group %s\r\n", group->name);
 	    else if (n == 1)	/* From: */
-		printf("Leafnode <news@%s>\r\n", fqdn);
+		printf("Leafnode <%s>\r\n", newsadmin);
 	    else if (n == 2)	/* Date: */
 		printf("%s\r\n", rfctime());
 	    else if (n == 3)	/* Message-ID: */
@@ -1544,56 +1488,7 @@ doxhdr(char *arg)
     return;
 }
 
-static int
-parserange(char *arg, unsigned long *a, unsigned long *b)
-{
-    char *l = arg;
-    /* no argument */
-    if (!*l) return 0;
-
-    /* parse */
-    if (*l != '-')
-	*a = strtoul(arg, &l, 10);
-    SKIPLWS(l);
-    if (*l == '-') {
-	++l;
-	SKIPLWS(l);
-	if (isdigit((unsigned char)*l))
-	    *b = strtoul(l, &l, 10);
-    } else {
-	*b = *a;
-    }
-    SKIPLWS(l);
-
-    /* trailing garbage */
-    if (l && *l) {
-	return 0;
-    }
-
-    return 1;
-}
-
-static int
-is_pseudogroup(/*@null@*/ const struct newsgroup *g)
-{
-    if (!g) return 0;
-    if (!chdirgroup(g->name, FALSE)) return 1;
-/*    if (isinteresting(g->name)) return 0; */
-    if (g->last < g->first) return 1;
-    if (g->last <= 1) return 1;
-    return 0;
-}
-
 static void
-nogroup(void)
-{
-    printf("412 Use the GROUP command first\r\n");
-    if (debugmode)
-	syslog(LOG_DEBUG, ">412 Use the GROUP command first");
-    return;
-}
-
-void
 doxover(char *arg)
 {
     unsigned long a, b, art;
@@ -1608,7 +1503,9 @@ doxover(char *arg)
     b = group->last;
     if (b < a) b = a;
 
-    if (!parserange(arg, &a, &b)) {
+    if (!arg || !*arg)
+	a = b = artno;
+    else if (!parserange(arg, &a, &b)) {
 	printf("502 Usage: OVER first[-[last]]\r\n");
 	if (debugmode)
 	    syslog(LOG_DEBUG, ">502 Usage: OVER first[-[last]]");
@@ -1657,20 +1554,18 @@ doxover(char *arg)
 		   "postings %lu-%lu:", group->first, group->first);
 	printf("%lu\t"
 	       "Leafnode placeholder for group %s\t"
-	       "news@%s (Leafnode)\t%s\t"
+	       "%s (Leafnode)\t%s\t"
 	       "<leafnode:placeholder:%s@%s>\t\t1000\t40\r\n", group->first,
-	       group->name, fqdn, rfctime(), group->name, fqdn);
+	       group->name, newsadmin, rfctime(), group->name, fqdn);
 	printf(".\r\n");
 	if (debugmode)
 	    syslog(LOG_DEBUG, ">%lu\tLeafnode placeholder for group %s\t"
-		   "news@%s (Leafnode)\t%s\t<leafnode:placeholder:%s@%s>\t\t1000\t40",
-		   group->first, group->name, fqdn, rfctime(), group->name, fqdn);
+		   "%s (Leafnode)\t%s\t<leafnode:placeholder:%s@%s>\t\t1000\t40",
+		   group->first, group->name, newsadmin, rfctime(), group->name, fqdn);
     }
 }
 
-
-
-void
+static void
 dolistgroup(const char *arg)
 {
     unsigned long art;
@@ -1719,7 +1614,118 @@ dolistgroup(const char *arg)
     printf(".\r\n");
 }
 
+static void
+parser(void)
+{
+    char *arg;
+    int n;
+    size_t size;
 
+    mgetaline_settimeout(timeout_client);
+
+    while ((cmd = mgetaline(stdin))) {
+	if (debug == 1)
+	    syslog(LOG_DEBUG, "<%s", cmd);
+
+	size = strlen(cmd);
+	if (size == 0)
+	    continue;		/* ignore */
+	if (size > MAXLINELENGTH || (long)size > (long)INT_MAX) {
+	    /* ignore attempts at buffer overflow */
+	    if (debugmode)
+		syslog(LOG_DEBUG, ">500 Dazed and confused");
+	    printf("500 Dazed and confused\r\n");
+	    continue;
+	}
+
+	/* parse command line */
+	n = 0;
+	while (isalpha((unsigned char)cmd[n]))
+	    n++;
+	while (isspace((unsigned char)cmd[n]))
+	    cmd[n++] = '\0';
+
+	arg = cmd + n;
+
+	while (cmd[n])
+	    n++;
+	n--;
+	while (n >= 0 && isspace((unsigned char)cmd[n]))
+	    cmd[n--] = '\0';
+
+	if (!strcasecmp(cmd, "quit")) {
+	    if (debugmode)
+		syslog(LOG_DEBUG, ">205 Always happy to serve!");
+	    printf("205 Always happy to serve!\r\n");
+	    return;
+	}
+	rereadactive();
+	if (!strcasecmp(cmd, "article")) {
+	    doarticle(arg, 3);
+	} else if (!strcasecmp(cmd, "head")) {
+	    doarticle(arg, 2);
+	} else if (!strcasecmp(cmd, "body")) {
+	    doarticle(arg, 1);
+	} else if (!strcasecmp(cmd, "stat")) {
+	    doarticle(arg, 0);
+	} else if (!strcasecmp(cmd, "help")) {
+	    dohelp();
+	} else if (!strcasecmp(cmd, "last")) {
+	    domove(-1);
+	} else if (!strcasecmp(cmd, "next")) {
+	    domove(1);
+	} else if (!strcasecmp(cmd, "list")) {
+	    dolist(arg);
+	} else if (!strcasecmp(cmd, "date")) {
+	    dodate();
+	} else if (!strcasecmp(cmd, "mode")) {
+	    if (debugmode)
+		syslog(LOG_DEBUG, ">200 Leafnode %s, pleased to meet you!",
+		       version);
+	    printf("200 Leafnode %s, pleased to meet you!\r\n", version);
+	} else if (!strcasecmp(cmd, "newgroups")) {
+	    donewgroups(arg);
+	} else if (!strcasecmp(cmd, "newnews")) {
+	    if (debugmode)
+		syslog(LOG_DEBUG,
+		       ">500 NEWNEWS is meaningless for this server");
+	    printf("500 NEWNEWS is meaningless for this server\r\n");
+	} else if (!strcasecmp(cmd, "post")) {
+	    if (dopost()) break;
+	} else if (!strcasecmp(cmd, "slave")) {
+	    if (debugmode)
+		syslog(LOG_DEBUG, ">202 Cool - I always wanted a slave");
+	    printf("202 Cool - I always wanted a slave\r\n");
+	} else if (!strcasecmp(cmd, "xhdr")) {
+	    doxhdr(arg);
+	} else if (!strcasecmp(cmd, "hdr")) {
+	    doxhdr(arg);
+	} else if (!strcasecmp(cmd, "xover")) {
+	    doxover(arg);
+	} else if (!strcasecmp(cmd, "over")) {
+	    doxover(arg);
+	} else if (!strcasecmp(cmd, "listgroup")) {
+	    dolistgroup(arg);
+	} else if (!strcasecmp(cmd, "group")) {
+	    if (dogroup(arg)) break;
+	} else {
+	    if (debugmode)
+		syslog(LOG_DEBUG, ">500 Unknown command");
+	    printf("500 Unknown command\r\n");
+	}
+	if (ferror(stdout) || fflush(stdout)) {
+	    syslog(LOG_ERR, "Cannot write to client.");
+	    break;
+	}
+    }
+    if (debugmode)
+	syslog(LOG_DEBUG, "Client timeout, disconnecting.");
+
+    /* There was once a 400 error message here. It confused broken
+     * clients, most notably, tin.
+     * Future NNTP drafts command that we don't send stuff back on
+     * timeout, so we anticipate these. */
+}
 
 int
 main(int argc, char **argv)
@@ -1744,6 +1750,7 @@ main(int argc, char **argv)
     char ownip[ADDRLEN]; /* RATS: ignore */
     char origfqdn[FQDNLEN + 1]; /* RATS: ignore */
 
+    ln_log_use_console(0); /* disable console logging */
     (void)argc;	/* quiet compiler warning */
     myopenlog("leafnode");
 
@@ -1855,7 +1862,11 @@ main(int argc, char **argv)
     rereadactive();
 
     parser();
+
     (void)fflush(stdout);
+    freeactive(active);
+    freexover();
+    freeconfig();
     sleep(1); /* protect against process ID induced file name collisions */
     exit(0);
 }
