@@ -363,8 +363,8 @@ dogroup(/*@null@*/ struct newsgroup *g, const char *name, int expdays)
 					|| link(de->d_name, t)))
 			    {
 				    ln_log(LNLOG_SERR, LNLOG_CGROUP,
-					   "%s: relink of %s <-> %s failed: %m (%s)",
-					   name, p, de->d_name, t);
+					   "%s: relink of %s <-> %s failed: %s (%s)",
+					   name, p, de->d_name, strerror(errno), t);
 			    } else {
 				ln_log(LNLOG_SINFO, LNLOG_CARTICLE,
 					"%s: relinked message %s <-> %s", name, p, de->d_name);
